﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using CloneMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItem>;
using GeneticExchangeMapType = HeuristicLab.Core.ItemList<HeuristicLab.Core.IItem>;
using TraceMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItemList<HeuristicLab.Core.IItem>>;

namespace HeuristicLab.EvolutionaryTracking {
  /// <summary>
  /// An operator which performs some cleanup at the end of the main loop.
  /// </summary>
  [Item("CleanupOperator", "This operator should be the last to run in the main loop and clean up the tracking-related data structures.")]
  [StorableClass]
  public sealed class CleanupOperator : SingleSuccessorOperator {
    #region Parameter Names
    private const string GlobalTraceMapParameterName = "GlobalTraceMap";
    private const string GlobalCloneMapParameterName = "GlobalCloneMap";
    private const string GlobalFragmentMapParameterName = "GlobalFragmentMap";
    private const string GlobalGeneticExchangeMapParameterName = "GlobalGeneticExchangeMap";
    private const string PopulationGraphParameterName = "PopulationGraph";
    private const string GenerationsParameterName = "Generations";
    private const string MaximumGenerationsParameterName = "MaximumGenerations";
    private const string ResultsParameterName = "Results";
    private const string CleanGenealogyGraphParameterName = "CleanGenealogyGraph";
    private const string UpdateGenealogyGraphParameterName = "UpdateGenealogyGraph";
    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";
    private const string SymbolicExpressionTreeQualityParameterName = "Quality";
    #endregion

    #region Parameter Properties
    public LookupParameter<TraceMapType> GlobalTraceMapParameter {
      get { return (LookupParameter<TraceMapType>)Parameters[GlobalTraceMapParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalCloneMapParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalFragmentMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalFragmentMapParameterName]; }
    }
    public LookupParameter<GeneticExchangeMapType> GlobalGeneticExchangeMapParameter {
      get { return (LookupParameter<GeneticExchangeMapType>)Parameters[GlobalGeneticExchangeMapParameterName]; }
    }
    public LookupParameter<IntValue> GenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public LookupParameter<IntValue> MaximumGenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters[MaximumGenerationsParameterName]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ValueParameter<BoolValue> CleanGenealogyGraphParameter {
      get { return (ValueParameter<BoolValue>)Parameters[CleanGenealogyGraphParameterName]; }
    }
    public ValueParameter<BoolValue> UpdateGenealogyGraphParameter {
      get { return (ValueParameter<BoolValue>)Parameters[UpdateGenealogyGraphParameterName]; }
    }
    public IScopeTreeLookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (IScopeTreeLookupParameter<ISymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> SymbolicExpressionTreeQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters[SymbolicExpressionTreeQualityParameterName]; }
    }
    #endregion

    #region Properties
    private CloneMapType GlobalCloneMap { get { return GlobalCloneMapParameter.ActualValue; } }
    private TraceMapType GlobalTraceMap { get { return GlobalTraceMapParameter.ActualValue; } }
    private CloneMapType GlobalFragmentMap { get { return GlobalFragmentMapParameter.ActualValue; } }
    private GeneticExchangeMapType GlobalGeneticExchangeMap { get { return GlobalGeneticExchangeMapParameter.ActualValue; } }
    private IntValue Generations { get { return GenerationsParameter.ActualValue; } }
    private IntValue MaximumGenerations { get { return MaximumGenerationsParameter.ActualValue; } }
    private ResultCollection Results { get { return ResultsParameter.ActualValue; } }
    public BoolValue CleanGenealogyGraph { get { return CleanGenealogyGraphParameter.Value; } }
    public BoolValue UpdateGenealogyGraph { get { return UpdateGenealogyGraphParameter.Value; } }
    private IEnumerable<DoubleValue> Quality { get { return SymbolicExpressionTreeQualityParameter.ActualValue; } }
    private IEnumerable<ISymbolicExpressionTree> SymbolicExpressionTree { get { return SymbolicExpressionTreeParameter.ActualValue; } }
    #endregion

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey(GlobalGeneticExchangeMapParameterName))
        Parameters.Add(new LookupParameter<GeneticExchangeMapType>(GlobalGeneticExchangeMapParameterName));
      if (!Parameters.ContainsKey(CleanGenealogyGraphParameterName))
        Parameters.Add(new ValueParameter<BoolValue>(CleanGenealogyGraphParameterName));
      if (!Parameters.ContainsKey(UpdateGenealogyGraphParameterName))
        Parameters.Add(new ValueParameter<BoolValue>(UpdateGenealogyGraphParameterName));
      if (!Parameters.ContainsKey(SymbolicExpressionTreeParameterName))
        Parameters.Add(new ScopeTreeLookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression trees to analyze."));
      if (!Parameters.ContainsKey(SymbolicExpressionTreeQualityParameterName))
        Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>(SymbolicExpressionTreeQualityParameterName, "The qualities of the symbolic expression trees"));
    }

    [StorableConstructor]
    private CleanupOperator(bool deserializing) : base(deserializing) { }
    private CleanupOperator(SingleSuccessorOperator original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) { return new CleanupOperator(this, cloner); }

    public CleanupOperator()
      : base() {
      Parameters.Add(new LookupParameter<TraceMapType>(GlobalTraceMapParameterName));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalCloneMapParameterName));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalFragmentMapParameterName));
      Parameters.Add(new LookupParameter<GeneticExchangeMapType>(GlobalGeneticExchangeMapParameterName));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName));
      Parameters.Add(new LookupParameter<IntValue>(MaximumGenerationsParameterName));
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new ValueParameter<BoolValue>(CleanGenealogyGraphParameterName));
      Parameters.Add(new ValueParameter<BoolValue>(UpdateGenealogyGraphParameterName));
      Parameters.Add(new ScopeTreeLookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression trees to analyze."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>(SymbolicExpressionTreeQualityParameterName, "The qualities of the symbolic expression trees"));
    }

    public override IOperation Apply() {
      if (GlobalCloneMap != null) GlobalCloneMap.Clear();
      if (GlobalFragmentMap != null) GlobalFragmentMap.Clear();
      if (GlobalTraceMap != null) GlobalTraceMap.Clear();
      if (GlobalGeneticExchangeMap != null) GlobalGeneticExchangeMap.Clear();

      if (!Results.ContainsKey(PopulationGraphParameterName)) return base.Apply();
      var graph = (SymbolicExpressionTreeGenealogyGraph)Results[PopulationGraphParameterName].Value;
      if (graph == null) return base.Apply();

      if (UpdateGenealogyGraph.Value) {
        // update vertex qualities if the trees in the scope were optimized by the ConstantOptimizationAnalyzer
        ISymbolicExpressionTree[] trees = SymbolicExpressionTree.ToArray();
        double[] qualities = Quality.Select(x => x.Value).ToArray();

        for (int i = 0; i != trees.Length; ++i) {
          foreach (var graphNode in graph.GetGraphNodes(trees[i])) {
            graphNode.Quality = qualities[i];
          }
        }
      }

      // if we are at the end of the run, then we clear the genealogy graph (to save memory)
      if (CleanGenealogyGraph.Value) {
        if (Generations.Value == MaximumGenerations.Value)
          if (Results.ContainsKey(PopulationGraphParameterName)) {
            Results.Remove(PopulationGraphParameterName);
            graph.Dispose();
          }
      }
      return base.Apply();
    }
  }
}
